/* Check book program - version 5 */
/* See the accompanying readme.txt file for more detailed comments */

/*
 * Include system header files
 *
 * these header files contain definitions and prototypes
 * that allow you to call various system supplied routines
 * (e.g., printf, gets, atoi, atof, etc.)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * Data Structures
 *
 * We had to move this definition before the prototypes
 * since some of the routines now use it as an argument
 */

struct xact {
       char *x_type;
       int x_num;
       char *x_date;
       char *x_desc;
       float x_amt;
       float x_bal;
};

/*
 * Function Prototypes
 *
 * These lines tell the compiler what type of routines we
 * will be defining and using
 */

int main(void);
void do_charges(void);
void do_checks(void);
void do_deposits(void);
void do_mac(void);
void set_balance(void);
void store_xact(char *type, int num, char *date, char *desc, float amt, float bal);
void display_history(void);
void load_file(char *fname);
void save_file(char *fname);
void save_xact(struct xact *x, FILE *fp);
int load_xacts(struct xact *x, FILE *fp);
void strinp_crlf(char *s);


/*
 * Global Variables
 * The following variables can be seen and used by all the routines
 * in this program
 */

struct xact Xacts[100];
int num_xacts=0, max_xacts = 100;

float balance=0.0;
char filename[BUFSIZ];

/*
 * The Main Routine
 *
 * This routine is called when the program first starts.
 * What it does is determined by you, and its actions control
 * what the program does
 */

int main()
{
    /*
     * Local Variables
     *
     * These variables can be seen and used only by this routine
     */

   int opt;
   char buf[BUFSIZ];

   /*
    * Executable Code
    *
    * This is the start of the code that actually runs
    *
    * Typically, this is where you do one-time only
    * type operations that need to be done before the
    * program starts.
    */

    puts("\nWelcome to the Super Whizbang Checking Program!");

    /* see if user wants to load an existing checkbook file */
    printf("Enter file name to load, or ENTER for a new book: ");
    gets(buf);
    strcpy(filename, buf);      /* save file name for exit */
    /* if user entered a file name read file into Xacts array */
    if (filename[0] != '\0') load_file(buf);

    /*
     * Main Loop
     *
     * This loop displays a menu for the user, prompts for an option,
     * and then performs the task requested by the user.
     *
     * It will keep doing this over and over again until the user
     * asks to quit (option 9).
     */

   do
   { /* start of main loop */

    /* display the current balance */
      printf("Your balance is now %.2lf\n", balance);

      /* display a menu of choices for the user */
      puts("\nOptions are:");
      puts("1. Set balance");
      puts("2. Process checks");
      puts("3. Process deposits");
      puts("4. Process MAC transactions");
      puts("5. Process service charges");
      puts("6. Display History");
      puts("9. Exit");
      printf("Enter option: "); /* ask for option from the user */

      /* get the user's option and process it */
      gets(buf); /* get a line from the user */
      opt = atoi(buf);  /* convert the line into its numeric value */
      switch(opt)
      { /* branch based on the user's option */
         case 1: set_balance(); break;
         case 2: do_checks(); break;
         case 3: do_deposits(); break;
         case 4: do_mac(); break;
         case 5: do_charges(); break;
         case 6: display_history(); break;
         case 9: puts("Bye!"); break;
         default: puts("Invalid option - try again");
      }
   } while (opt != 9); /* end of the main loop */

   /*
    * if we get here, then the user asked to quit the program
    *
    * At this point, we should do any clean up that is needed
    * and then leave in an orderly fashion
    */

    /* see if user wants to save the current checkbook */
   printf("Save to file? (y or n): ");
   gets(buf);
   if (buf[0] != 'n')
   { /* saving to file - now get file name */
        if (filename[0] == '\0')
        { /* no file was loaded - make a default name */
            strcpy(filename, "chkbook.txt");
        }
        printf("Enter file name, or ENTER for %s: ", filename);
        gets(buf);
        if (buf[0] != '\0') strcpy(filename, buf);  /* don't use default */
        save_file(filename);    /* save Xacts to file */
   }

   puts("Thanks for using Super Whizbang Checking!  Come again!");

   return 0;    /* normal return */
}

/* the following routines handle the various menu options */

/*
 * 					**** Do Charges ****
 *
 * This routine handles service charges and other misc. charges
 * to the account.
 *
 * The routine prompts the user for necessary information,
 * displays a line representing the transaction, and updates
 * the account balance.
 *
 * When reading responses from the user, we use an input buffer
 * that is plenty big.  We then make sure that the user didn't
 * type too many characters.  If we didn't do that, we risk
 * corrupting the data and causing errors
 *
 */
void do_charges(void)
{
   char buf[BUFSIZ];                /* input buffer */
   char chg_date[9], chg_desc[21];  /* date and description */
   float chg_amt;                   /* amount of charge */

   /* get the date of the charge */
   printf("Enter date of charge (MM/DD/YY): ");
   gets(buf);
   buf[8] = '\0';   /* make sure user didn't type too much */
   strcpy(chg_date, buf);   /* copy data to correct field */

   /* now get a description of the charge */
   printf("Enter description of charge: ");
   gets(buf);
   buf[20] = '\0';
   strcpy(chg_desc, buf); /* copy data to description field */

   /* now handle the amount of the charge */
   printf("Enter amount of charge: ");
   gets(buf);
   chg_amt = atof(buf); /* convert amount to floating point */

   /* display a line showing the transaction */
   printf("%-8s %-20s %10.2f\n", chg_date, chg_desc, chg_amt);

   /* update the balance by subtracting the amount from the balance */
   balance -= chg_amt;

   store_xact("Chg", 0, chg_date, chg_desc, chg_amt, balance);
}

void do_checks(void)
{
   char buf[BUFSIZ], chk_date[9], chk_desc[21];
   int chk_no;
   float chk_amt;

   printf("Enter check number: ");
   gets(buf);
   chk_no = atoi(buf);  /* convert to integer */

   printf("Enter date of check: ");
   gets(buf);
   buf[8] = 0;
   strcpy(chk_date, buf);

   printf("Enter desciption of check: ");
   gets(buf);
   buf[20] = 0;
   strcpy(chk_desc, buf);

   printf("Enter amount of check: ");
   gets(buf);
   chk_amt = atof(buf);

   printf("%4d %-8s %-20s %10.2f\n", chk_no, chk_date, chk_desc, chk_amt);

   balance -= chk_amt;

   store_xact("Chk", chk_no, chk_date, chk_desc, chk_amt, balance);
}

void do_deposits(void)
{
   char buf[BUFSIZ], dep_date[8], dep_desc[20];
   float dep_amt;

   printf("Enter date of deposit: ");
   gets(buf);
   buf[8] = 0;
   strcpy(dep_date, buf);

   printf("Enter description of deposit: ");
   gets(buf);
   buf[20] = 0;
   strcpy(dep_desc, buf);

   printf("Enter amount of deposit: ");
   gets(buf);
   dep_amt = atof(buf);

   printf("%-8s %-20s %10.2f\n", dep_date, dep_desc, dep_amt);

   balance += dep_amt;

   store_xact("Dep", 0, dep_date, dep_desc, dep_amt, balance);
}

void do_mac(void)
{
   char buf[BUFSIZ], mac_date[9], mac_desc[21], mac_xact;
   float mac_amt;
   int sign;

   /* keep looping and processing transactions until done */
   while (1)
   {
    /* display a brief menu for the user */
      printf("Your balance is now %.2f\n", balance);
      printf("Deposit, Withdrawl, or quit? (d, w, q): ");
      gets(buf);        /* get a response from the user */
      mac_xact = buf[0]; /* get the first letter of the response */
      switch(mac_xact)
      { /* now branch to appropriate location */
        /* this next line converts lower case w to upper case W */
         case 'w': mac_xact = 'W';
         case 'W':
              sign = -1;    /* subtract amount from total*/
              break;
         case 'd': mac_xact = 'D';
         case 'D':
              sign = 1;     /* add amount to total */
              break;
         case 'q':
         case 'Q':
              return;       /* quit the loop */
         default:
                puts("Invalid transaction.  Please try again.");
                continue;
      }

      /* doing a transaction - handle the common fields */

      /* get date of transaction */
      printf("Enter date of transaction: ");
      gets(buf);
      buf[8] = 0;
      strcpy(mac_date, buf);

      /* get description of transaction */
      printf("Enter description: ");
      gets(buf);
      buf[20] = 0;
      strcpy(mac_desc, buf);

      /* get amount of transaction */
      printf("Enter amount: ");
      gets(buf);
      mac_amt = atof(buf);

      /* display line representing the transaction */
      sprintf(buf, "MAC%c", mac_xact);
      printf("%-8s %s %-20s %10.2f\n",
             mac_date, buf, mac_desc, mac_amt);

      /* update balance according to sign */
      if (sign == 1) balance += mac_amt; else balance -= mac_amt;
   	  store_xact(buf, 0, mac_date, mac_desc, mac_amt, balance);
   }

}

/*
 * Set Balance
 *
 * This routine allows the user to set the starting balance
 * It would typically be used at the start of the program.
 * If we were able to remember what the closing balance was
 * the last time we ran, we wouldn't need this routine.
 */

void set_balance(void)
{
   char buf[BUFSIZ];

   printf("Enter new balance: "); /* ask for the new balance */
   gets(buf);           /* get a string from the user */
   balance = atof(buf); /* convert string the user typed into a number */
   /*
    * Since balance is a global variable, this change is permanent.
    *
    * If we were to define a local variable in this routine called
    * balance, then when we left this routine, the global balance
    * would be unchanged
    */
}

/*
 * Store Xact
 *
 * This routine stores the transaction into a table
 *
 * This table is used to print history.
 *
 * We can also use this in a later release to store transactions
 * for later use.
 *
 */
void store_xact(char *type, int num, char *date, char *desc, float amt, float bal)
{
   if (num_xacts >= max_xacts)
   {
      puts("Transaction table too full!");
      return;
   }

   Xacts[num_xacts].x_type = strdup(type);
   Xacts[num_xacts].x_num = num;
   Xacts[num_xacts].x_date = strdup(date);
   Xacts[num_xacts].x_desc = strdup(desc);
   Xacts[num_xacts].x_amt = amt;
   Xacts[num_xacts].x_bal = bal;

   num_xacts++;
}

/*
 * Display History
 *
 * This routine will display a list of transactions
 *
 */

void display_history(void)
{
   int i;
   static char *fmt = "%-4s %5s %-8s %-20s %9.2f %9.2f\n";
   char num[6];
   char buf[BUFSIZ];

   for (i=0; i < num_xacts; i++)
   {
      if (strcmp(Xacts[i].x_type, "Chk") == 0)
         sprintf(num, "%5d", Xacts[i].x_num);
      else num[0] = '\0';

      printf(fmt, Xacts[i].x_type, num, Xacts[i].x_date,
                  Xacts[i].x_desc, Xacts[i].x_amt,
                  Xacts[i].x_bal);

      if (i > 0 && (i % 20) == 0)
      {
         puts("Press ENTER to continue ...");
         gets(buf);
      }
   }
   puts("Press ENTER to continue ...");
   gets(buf);
}

void save_file(char *fname)
{
    int i;
    FILE *fp;

    if ((fp = fopen(fname, "w")) == NULL)
    { /* error when opening file */
        fprintf(stderr, "Can't save to file (%s)\n", fname);
        return;
    }

    for (i=0; i < num_xacts; i++) save_xact(Xacts + i, fp);

    fclose(fp);
}

void save_xact(struct xact *x, FILE *fp)
{ /* save transaction to file */

       /*
        * write out this transaction to the specified file
        *
        * to make it easier to read back in, each field
        * will be written to a separate line.
        *
        * to make it easier to look at using a text editor,
        * we include a line with equal signs after the record
        *
        * a *real* program would have the entire transaction on
        * a single line and then parse the line into fields
        *
        */

       fprintf(fp, "%s\n%d\n%s\n%s\n%f\n%f\n======================\n",
       	x->x_type, x->x_num, x->x_date, x->x_desc, x->x_amt, x->x_bal);
}

void load_file(char *fname)
{
    FILE *fp;
    int nxact;  /* how many transactions we read */

    if ((fp = fopen(fname, "r")) == NULL)
    { /* error opening file */
        fprintf(stderr, "Can't open file for reading (%s)\n", fname);
        return;
    }

    num_xacts = 0;      /* reset number of transactions */
    balance = 0;        /* reset account balance */
    /*
   	 * read as many transactions as we can
     * we should pass a maximum number of transactions
     * to this routine to make sure we don't overrun
     * the array
     *
     * Try to add that logic yourself.
     */
    nxact = load_xacts(Xacts, fp);

    if (nxact > 0)
    { /* some transactions loaded - get last balance for current balance */
        num_xacts = nxact;
        balance = Xacts[num_xacts-1].x_bal;
    }

    fclose(fp);
}

int load_xacts(struct xact *x, FILE *fp)
{
    int nrecs = 0;      /* number of records actually read */
    char buf[BUFSIZ];   /* input buffer */

    /*
     * We loop reading records
     *
     * Read a line and place it in the current record
     * When the record has been read, increment counter & pointer
     *
     * When we hit an end-of-file, fgets will return NULL
     *
     * We should get an end-of-file only on the first line of
     * a transaction.
     *
     * We should be more careful copying data into the table
     * we should check the length of each string before we copy
     * we should also test for end-of-file after each read
     * Try to add that logic yourself
     *
     */

     while (fgets(buf, sizeof(buf), fp) != NULL)
     { /* xact type read - we have another record */
        strinp_crlf(buf);    /* strip CR-LF characters from line */
        x->x_type = strdup(buf);

        /* check number */
        fgets(buf, sizeof(buf), fp);
        x->x_num = atoi(buf);   /* convert to number */

        /* date */
        fgets(buf, sizeof(buf), fp);
        strinp_crlf(buf);
        x->x_date = strdup(buf);

        /* description */
        fgets(buf, sizeof(buf), fp);
        strinp_crlf(buf);
        x->x_desc = strdup(buf);

        /* amount */
        fgets(buf, sizeof(buf), fp);
        x->x_amt = atof(buf);

        /* balance */
        fgets(buf, sizeof(buf), fp);
        x->x_bal = atof(buf);

        /* ignore the line with all equal signs */
        fgets(buf, sizeof(buf), fp);

        /* record read - increment counter & record */
        nrecs++;    /* increment counter */
        x++;        /* point to next record */
     }
     return nrecs;
}

void strinp_crlf(char *s)
{ /* remove carriage-return and/or line-feed from end of line */
    char *p;    /* pointer used for searching */

    /* find end of line */
    for (p=s; *p != '\0' && *p != '\r' && *p != '\n'; p++);
    *p = '\0';  /* truncate line */
}

